/*
 * Copyright (C) 2012  Alejandro Mery <amery@geeks.cl>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/string.h>

#include <asm/setup.h>
#include <asm/sizes.h>
#include <asm/mach-types.h>

#include <plat/core.h>
#include <plat/platform.h>
#include <plat/system.h>

struct brom_header {
	u32 jump_instruction;	/* one intruction jumping to real code */
	char magic[8];		/* ="eGON.BRM",  not C-style string. */
	u32 length;		/* generated by PC */
	char Boot_vsn[4];	/* Boot version */
	char eGON_vsn[4];	/* eGON version */
	char platform[8];	/* platform information */
};

static inline const struct brom_header *_sunxi_brom(void)
{
	const struct brom_header *brom = (void*)SW_VA_BROM_BASE + 0x4000;
	if (strncmp(brom->magic, "eGON.BRM", 8) == 0)
		return brom;
	else
		return NULL;
}

int sunxi_pr_brom(void)
{
	const struct brom_header *brom = _sunxi_brom();
	if (brom) {
		pr_info("BROM Ver: %.4s %.4s %.8s\n",
			brom->Boot_vsn,
			brom->eGON_vsn,
			brom->platform);
		return 1;
	}
	pr_err("SUNXI BROM not found");
	return 0;
}

u32 sunxi_chip_id(void)
{
	const struct brom_header *brom = _sunxi_brom();
	if (brom) {
		char *endptr;
		u32 chip_id = simple_strtoul(brom->platform, &endptr, 10);
		if (*endptr == '\0')
			return chip_id;

		pr_err("BROM: failed to parse chip-id: \"%s\"\n",
		       brom->platform);
	}
	return 0;
}

int sunxi_pr_chip_id()
{
	u32 chip_id = sunxi_chip_id();
	const char *name;

	switch (chip_id) {
	case SUNXI_CHIP_ID_A10:
		name = "A10"; break;
	case SUNXI_CHIP_ID_A13:
		name = "A13"; break;
	default:
		name = "Unknown";
	}

	if (machine_is_sun4i())
		pr_info("chip-id: %s (AW%u revision %c)\n", name,
			chip_id, 'A' + sw_get_ic_ver());
	else
		pr_info("chip-id: %s (AW%u)\n", name, chip_id);

	return (chip_id != 0);
}

#ifdef CONFIG_SUNXI_MALI_RESERVED_MEM
void __init sunxi_mali_core_fixup(struct tag *t, char **cmdline,
				  struct meminfo *mi)
{
	u32 bank = 0;
	for (; t->hdr.size; t = tag_next(t)) if (t->hdr.tag == ATAG_MEM) {
		if (bank) {
			mi->nr_banks++;
			mi->bank[bank].start = t->u.mem.start;
			mi->bank[bank].size = t->u.mem.size;
		} else { /* first bank */
			u32 size = t->u.mem.size / SZ_1M;
			mi->nr_banks = 1;
			mi->bank[0].start = t->u.mem.start;
			if (size < 512) {
				mi->bank[0].size = SZ_1M * size;

				pr_err("MALI: not enough memory in first bank to make reserve.\n");
			} else {
				mi->bank[0].size = SZ_1M * (512 - 64);
				size -= 512;
				if (size) {
					bank++;
					mi->nr_banks++;
					mi->bank[1].start = t->u.mem.start + (512 * SZ_1M);
					mi->bank[1].size = SZ_1M * size;
				}

				pr_info("Memory cut off:\n");
				pr_reserve_info("MALI", t->u.mem.start + SZ_512M - SZ_64M,
						SZ_64M);
			}
		}
		bank++;
	}
}
#endif
